"""This module contains constructor functions for the primitive configurable parameter types.

The available parameter types are: `configurable_integer`, `configurable_float`, `configurable_boolean`,
`string_selectable` and `float_selectable`.
"""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


from typing import List, Optional, TypeVar, Union

import attr

from otx.api.configuration.enums import AutoHPOState, ConfigElementType, ModelLifecycle
from otx.api.configuration.ui_rules import NullUIRules, UIRules

from .configurable_enum import ConfigurableEnum
from .metadata_keys import (
    AFFECTS_OUTCOME_OF,
    AUTO_HPO_STATE,
    AUTO_HPO_VALUE,
    DEFAULT_VALUE,
    DESCRIPTION,
    EDITABLE,
    HEADER,
    MAX_VALUE,
    MIN_VALUE,
    OPTIONS,
    TYPE,
    UI_RULES,
    VISIBLE_IN_UI,
    WARNING,
)
from .utils import (
    attr_strict_float_converter,
    attr_strict_float_on_setattr,
    attr_strict_int_validator,
    construct_attr_enum_selectable_converter,
    construct_attr_enum_selectable_onsetattr,
    construct_attr_selectable_validator,
    construct_attr_value_validator,
)

# pylint:disable=too-many-arguments

_ConfigurableEnum = TypeVar("_ConfigurableEnum", bound=ConfigurableEnum)


def set_common_metadata(
    default_value: Union[int, float, str, bool, ConfigurableEnum],
    header: str,
    description: str,
    warning: Optional[str],
    editable: bool,
    affects_outcome_of: ModelLifecycle,
    ui_rules: UIRules,
    visible_in_ui: bool,
    parameter_type: ConfigElementType,
    auto_hpo_state: AutoHPOState,
    auto_hpo_value: Optional[Union[int, float, str, bool, ConfigurableEnum]],
) -> dict:
    """Function to construct the dictionary of metadata that is common for all parameter types."""
    metadata = {
        DEFAULT_VALUE: default_value,
        DESCRIPTION: description,
        HEADER: header,
        WARNING: warning,
        EDITABLE: editable,
        VISIBLE_IN_UI: visible_in_ui,
        AFFECTS_OUTCOME_OF: affects_outcome_of,
        UI_RULES: ui_rules,
        TYPE: parameter_type,
        AUTO_HPO_STATE: auto_hpo_state,
        AUTO_HPO_VALUE: auto_hpo_value,
    }
    return metadata


def configurable_integer(
    default_value: int,
    header: str,
    min_value: int = 0,
    max_value: int = 255,
    description: str = "Default integer description",
    warning: str = None,
    editable: bool = True,
    visible_in_ui: bool = True,
    affects_outcome_of: ModelLifecycle = ModelLifecycle.NONE,
    ui_rules: UIRules = NullUIRules(),
    auto_hpo_state: AutoHPOState = AutoHPOState.NOT_POSSIBLE,
    auto_hpo_value: Optional[int] = None,
) -> int:
    """Constructs a configurable integer attribute, with the appropriate metadata.

    Args:
        default_value: integer to use as default for the parameter
        header: User friendly name for the parameter, which will be
            shown in the UI
        min_value: lower bound of the range of values this parameter can
            take. Defaults to 0
        max_value: upper bound of the range of values this parameter can
            take Defaults to 255
        description: A user friendly description of what this parameter
            does, what does it represent and what are the effects of
            changing it?
        warning: An optional warning message to caution users when
            changing this parameter. This message will be displayed in
            the UI. For example, for the parameter batch_size:
            `Increasing batch size increases GPU memory demands and may
            result in out of memory errors. Please update batch size
            with caution.`
        editable: Set to False to prevent the parameter from being
            edited in the UI. It can still be edited through the REST
            API or the SDK. Defaults to True
        visible_in_ui: Set to False to hide the parameter from the UI
            and the REST API. It will still be visible through the SDK.
            Defaults to True
        affects_outcome_of: Describes the stage of the ModelLifecycle in
            which this parameter modifies the outcome. See the
            documentation for the ModelLifecycle Enum for further
            details
        ui_rules: Set of rules to control UI behavior for this
            parameter. For example, the parameter can be shown or hidden
            from the UI based on the value of other parameters in the
            configuration. Have a look at the UIRules class for more
            details. Defaults to NullUIRules.
        auto_hpo_state: This flag reflects whether the parameter can be
            (or has been) optimized through automatic hyper parameter
            tuning (auto-HPO)
        auto_hpo_value: If auto-HPO has been executed for this
            parameter, this field will hold the optimized value for the
            configurable integer

    Returns:
        attrs Attribute of type `int`, with its metadata set according
        to the inputs
    """
    metadata = set_common_metadata(
        default_value=default_value,
        header=header,
        description=description,
        warning=warning,
        editable=editable,
        visible_in_ui=visible_in_ui,
        ui_rules=ui_rules,
        affects_outcome_of=affects_outcome_of,
        parameter_type=ConfigElementType.INTEGER,
        auto_hpo_state=auto_hpo_state,
        auto_hpo_value=auto_hpo_value,
    )

    metadata.update({MIN_VALUE: min_value, MAX_VALUE: max_value})
    value_validator = construct_attr_value_validator(min_value, max_value)
    type_validator = attr_strict_int_validator

    return attr.ib(
        default=default_value,
        type=int,
        validator=[value_validator, type_validator],
        on_setattr=attr.setters.validate,
        metadata=metadata,
    )


def configurable_float(
    default_value: float,
    header: str,
    min_value: float = 0.0,
    max_value: float = 255.0,
    description: str = "Default float description",
    warning: str = None,
    editable: bool = True,
    visible_in_ui: bool = True,
    affects_outcome_of: ModelLifecycle = ModelLifecycle.NONE,
    ui_rules: UIRules = NullUIRules(),
    auto_hpo_state: AutoHPOState = AutoHPOState.NOT_POSSIBLE,
    auto_hpo_value: Optional[float] = None,
) -> float:
    """Constructs a configurable float attribute, with the appropriate metadata.

    Args:
        default_value: float to use as default for the parameter
        header: User friendly name for the parameter, which will be
            shown in the UI
        min_value: lower bound of the range of values this parameter can
            take. Defaults to 0.0
        max_value: upper bound of the range of values this parameter can
            take Defaults to 255.0
        description: A user friendly description of what this parameter
            does, what does it represent and what are the effects of
            changing it?
        warning: An optional warning message to caution users when
            changing this parameter. This message will be displayed in
            the UI. For example, for the parameter batch_size:
            `Increasing batch size increases GPU memory demands and may
            result in out of memory errors. Please update batch size
            with caution.`
        editable: Set to False to prevent the parameter from being
            edited in the UI. It can still be edited through the REST
            API or the SDK. Defaults to True
        visible_in_ui: Set to False to hide the parameter from the UI
            and the REST API. It will still be visible through the SDK.
            Defaults to True
        affects_outcome_of: Describes the stage of the ModelLifecycle in
            which this parameter modifies the outcome. See the
            documentation for the ModelLifecycle Enum for further
            details
        ui_rules: Set of rules to control UI behavior for this
            parameter. For example, the parameter can be shown or hidden
            from the UI based on the value of other parameters in the
            configuration. Have a look at the UIRules class for more
            details. Defaults to NullUIRules.
        auto_hpo_state: This flag reflects whether the parameter can be
            (or has been) optimized through automatic hyper parameter
            tuning (auto-HPO)
        auto_hpo_value: If auto-HPO has been executed for this
            parameter, this field will hold the optimized value for the
            configurable float

    Returns:
        attrs Attribute of type `float`, with its metadata set according
        to the inputs
    """
    metadata = set_common_metadata(
        default_value=default_value,
        header=header,
        description=description,
        warning=warning,
        editable=editable,
        visible_in_ui=visible_in_ui,
        ui_rules=ui_rules,
        affects_outcome_of=affects_outcome_of,
        parameter_type=ConfigElementType.FLOAT,
        auto_hpo_state=auto_hpo_state,
        auto_hpo_value=auto_hpo_value,
    )

    metadata.update({MIN_VALUE: min_value, MAX_VALUE: max_value})
    value_validator = construct_attr_value_validator(min_value, max_value)
    type_validator = attr_strict_float_on_setattr

    return attr.ib(
        default=default_value,
        type=float,
        validator=[value_validator, type_validator],
        converter=attr_strict_float_converter,
        on_setattr=[attr.setters.convert, attr.setters.validate],
        metadata=metadata,
    )


def configurable_boolean(
    default_value: bool,
    header: str,
    description: str = "Default configurable boolean description",
    warning: str = None,
    editable: bool = True,
    visible_in_ui: bool = True,
    affects_outcome_of: ModelLifecycle = ModelLifecycle.NONE,
    ui_rules: UIRules = NullUIRules(),
    auto_hpo_state: AutoHPOState = AutoHPOState.NOT_POSSIBLE,
    auto_hpo_value: Optional[bool] = None,
) -> bool:
    """Constructs a configurable boolean attribute, with the appropriate metadata.

    Args:
        default_value: boolean to use as default for the parameter
        header: User friendly name for the parameter, which will be
            shown in the UI
        description: A user friendly description of what this parameter
            does, what does it represent and what are the effects of
            changing it?
        warning: An optional warning message to caution users when
            changing this parameter. This message will be displayed in
            the UI. For example, for the parameter batch_size:
            `Increasing batch size increases GPU memory demands and may
            result in out of memory errors. Please update batch size
            with caution.`
        editable: Set to False to prevent the parameter from being
            edited in the UI. It can still be edited through the REST
            API or the SDK. Defaults to True
        visible_in_ui: Set to False to hide the parameter from the UI
            and the REST API. It will still be visible through the SDK.
            Defaults to True
        affects_outcome_of: Describes the stage of the ModelLifecycle in
            which this parameter modifies the outcome. See the
            documentation for the ModelLifecycle Enum for further
            details
        ui_rules: Set of rules to control UI behavior for this
            parameter. For example, the parameter can be shown or hidden
            from the UI based on the value of other parameters in the
            configuration. Have a look at the UIRules class for more
            details. Defaults to NullUIRules.
        auto_hpo_state: This flag reflects whether the parameter can be
            (or has been) optimized through automatic hyper parameter
            tuning (auto-HPO)
        auto_hpo_value: If auto-HPO has been executed for this
            parameter, this field will hold the optimized value for the
            configurable boolean

    Returns:
        attrs Attribute of type `bool`, with its metadata set according
        to the inputs
    """
    metadata = set_common_metadata(
        default_value=default_value,
        header=header,
        description=description,
        warning=warning,
        editable=editable,
        visible_in_ui=visible_in_ui,
        ui_rules=ui_rules,
        affects_outcome_of=affects_outcome_of,
        parameter_type=ConfigElementType.BOOLEAN,
        auto_hpo_state=auto_hpo_state,
        auto_hpo_value=auto_hpo_value,
    )
    type_validator = attr.validators.instance_of(bool)

    return attr.ib(
        default=default_value,
        metadata=metadata,
        type=bool,
        validator=type_validator,
        on_setattr=attr.setters.validate,
    )


def float_selectable(
    default_value: float,
    header: str,
    options: List[float],
    description: str = "Default selectable description",
    warning: str = None,
    editable: bool = True,
    visible_in_ui: bool = True,
    affects_outcome_of: ModelLifecycle = ModelLifecycle.NONE,
    ui_rules: UIRules = NullUIRules(),
    auto_hpo_state: AutoHPOState = AutoHPOState.NOT_POSSIBLE,
    auto_hpo_value: Optional[float] = None,
) -> float:
    """Constructs a configurable float selectable attribute, with the appropriate metadata.

    Args:
        default_value: float to use as default for the parameter
        header: User friendly name for the parameter, which will be
            shown in the UI
        options: list of float options representing the values that this
            parameter can take
        description: A user friendly description of what this parameter
            does, what does it represent and what are the effects of
            changing it?
        warning: An optional warning message to caution users when
            changing this parameter. This message will be displayed in
            the UI. For example, for the parameter batch_size:
            `Increasing batch size increases GPU memory demands and may
            result in out of memory errors. Please update batch size
            with caution.`
        editable: Set to False to prevent the parameter from being
            edited in the UI. It can still be edited through the REST
            API or the SDK. Defaults to True
        visible_in_ui: Set to False to hide the parameter from the UI
            and the REST API. It will still be visible through the SDK.
            Defaults to True
        affects_outcome_of: Describes the stage of the ModelLifecycle in
            which this parameter modifies the outcome. See the
            documentation for the ModelLifecycle Enum for further
            details
        ui_rules: Set of rules to control UI behavior for this
            parameter. For example, the parameter can be shown or hidden
            from the UI based on the value of other parameters in the
            configuration. Have a look at the UIRules class for more
            details. Defaults to NullUIRules.
        auto_hpo_state: This flag reflects whether the parameter can be
            (or has been) optimized through automatic hyper parameter
            tuning (auto-HPO)
        auto_hpo_value: If auto-HPO has been executed for this
            parameter, this field will hold the optimized value for the
            float selectable

    Returns:
        attrs Attribute of type `float`, with its metadata set according
        to the inputs
    """
    metadata = set_common_metadata(
        default_value=default_value,
        header=header,
        description=description,
        warning=warning,
        editable=editable,
        visible_in_ui=visible_in_ui,
        ui_rules=ui_rules,
        affects_outcome_of=affects_outcome_of,
        parameter_type=ConfigElementType.FLOAT_SELECTABLE,
        auto_hpo_state=auto_hpo_state,
        auto_hpo_value=auto_hpo_value,
    )

    metadata.update({OPTIONS: options})
    value_validator = construct_attr_selectable_validator(options)
    type_validator = attr_strict_float_on_setattr

    return attr.ib(
        default=default_value,
        type=float,
        validator=[value_validator, type_validator],
        converter=attr_strict_float_converter,
        on_setattr=[attr.setters.convert, attr.setters.validate],
        metadata=metadata,
    )


def selectable(
    default_value: _ConfigurableEnum,
    header: str,
    description: str = "Default selectable description",
    warning: str = None,
    editable: bool = True,
    visible_in_ui: bool = True,
    affects_outcome_of: ModelLifecycle = ModelLifecycle.NONE,
    ui_rules: UIRules = NullUIRules(),
    auto_hpo_state: AutoHPOState = AutoHPOState.NOT_POSSIBLE,
    auto_hpo_value: Optional[str] = None,
) -> _ConfigurableEnum:
    """Constructs a selectable attribute from a pre-defined Enum, with the appropriate metadata.

    The list of options for display in the UI is inferred from the type of the ConfigurableEnum instance passed in
    as default_value.

    Args:
        default_value: OTXConfigurationEnum instance to use as default
            for the parameter
        header: User friendly name for the parameter, which will be
            shown in the UI
        description: A user friendly description of what this parameter
            does, what does it represent and what are the effects of
            changing it?
        warning: An optional warning message to caution users when
            changing this parameter. This message will be displayed in
            the UI. For example, for the parameter batch_size:
            `Increasing batch size increases GPU memory demands and may
            result in out of memory errors. Please update batch size
            with caution.`
        editable: Set to False to prevent the parameter from being
            edited in the UI. It can still be edited through the REST
            API or the SDK. Defaults to True
        visible_in_ui: Set to False to hide the parameter from the UI
            and the REST API. It will still be visible through the SDK.
            Defaults to True
        affects_outcome_of: Describes the stage of the ModelLifecycle in
            which this parameter modifies the outcome. See the
            documentation for the ModelLifecycle Enum for further
            details
        ui_rules: Set of rules to control UI behavior for this
            parameter. For example, the parameter can be shown or hidden
            from the UI based on the value of other parameters in the
            configuration. Have a look at the UIRules class for more
            details. Defaults to NullUIRules.
        auto_hpo_state: This flag reflects whether the parameter can be
            (or has been) optimized through automatic hyper parameter
            tuning (auto-HPO)
        auto_hpo_value: If auto-HPO has been executed for this
            parameter, this field will hold the optimized value for the
            string selectable

    Returns:
        attrs Attribute, with its type matching the type of
        `default_value`, and its metadata set according to the inputs
    """
    metadata = set_common_metadata(
        default_value=default_value,
        header=header,
        description=description,
        warning=warning,
        editable=editable,
        visible_in_ui=visible_in_ui,
        ui_rules=ui_rules,
        affects_outcome_of=affects_outcome_of,
        parameter_type=ConfigElementType.SELECTABLE,
        auto_hpo_state=auto_hpo_state,
        auto_hpo_value=auto_hpo_value,
    )

    metadata.update(default_value.get_class_info())

    type_validator = attr.validators.instance_of(ConfigurableEnum)
    value_validator = construct_attr_enum_selectable_onsetattr(default_value)

    # The Attribute returned by attr.ib is not compatible with the return typevar _ConfigurableEnum. However, as the
    # class containing the Attribute is instantiated the selectable type will correspond to the _ConfigurableEnum, so
    # mypy can ignore the error.
    return attr.ib(
        default=default_value,
        type=ConfigurableEnum,
        validator=[type_validator, value_validator],  # type: ignore
        converter=construct_attr_enum_selectable_converter(default_value),
        on_setattr=[attr.setters.convert, value_validator],
        metadata=metadata,
    )  # type: ignore


def string_attribute(value: str) -> str:
    """String attribute.

    Wrapper for attr.ib that can be used to overwrite simple string attributes in a class or parameter group
    definition.

    Args:
        value: string to be added as attribute

    Returns:
        attr.ib string attribute with its default value set to value
    """
    return attr.ib(default=value, type=str, kw_only=True)


def boolean_attribute(value: bool) -> bool:
    """Boolean attribute wrapper.

    Wrapper for attr.ib that can be used to overwrite simple boolean attributes in a class or parameter group
    definition.

    Args:
        value: boolean to be added as attribute

    Returns:
        attr.ib boolean attribute with its default value set to value
    """
    return attr.ib(default=value, type=bool, kw_only=True)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# flake8: noqa
import os

from .builder import build, build_workflow_hook
from .stage import Stage, get_available_types
from .version import __version__, get_version
from .workflow import Workflow


class MPAConstants:
    PACKAGE_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
    # PACKAGE_ROOT = os.path.dirname(Path(__file__).)
    RECIPES_PATH = os.path.join(PACKAGE_ROOT, "recipes")
    SAMPLES_PATH = os.path.join(PACKAGE_ROOT, "samples")
    MODELS_PATH = os.path.join(PACKAGE_ROOT, "models")


# print(f'pkg root ======> {MPAConstants.PACKAGE_ROOT}')

__all__ = [get_version, __version__, build, build_workflow_hook, Stage, get_available_types, Workflow, MPAConstants]



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import torch
from mmcv.runner import force_fp32
from mmdet.core import bbox_overlaps, multi_apply, reduce_mean
from mmdet.models.builder import HEADS
from mmdet.models.dense_heads.atss_head import ATSSHead

from otx.mpa.modules.models.heads.cross_dataset_detector_head import (
    CrossDatasetDetectorHead,
)
from otx.mpa.modules.models.losses.cross_focal_loss import CrossSigmoidFocalLoss

EPS = 1e-12


@HEADS.register_module()
class CustomATSSHead(CrossDatasetDetectorHead, ATSSHead):
    def __init__(
        self,
        *args,
        bg_loss_weight=-1.0,
        use_qfl=False,
        qfl_cfg=dict(
            type="QualityFocalLoss",
            use_sigmoid=True,
            beta=2.0,
            loss_weight=1.0,
        ),
        **kwargs
    ):
        if use_qfl:
            kwargs["loss_cls"] = qfl_cfg
        super().__init__(*args, **kwargs)
        self.bg_loss_weight = bg_loss_weight
        self.use_qfl = use_qfl

    @force_fp32(apply_to=("cls_scores", "bbox_preds", "centernesses"))
    def loss(self, cls_scores, bbox_preds, centernesses, gt_bboxes, gt_labels, img_metas, gt_bboxes_ignore=None):
        """Compute losses of the head.

        Args:
            cls_scores (list[Tensor]): Box scores for each scale level
                Has shape (N, num_anchors * num_classes, H, W)
            bbox_preds (list[Tensor]): Box energies / deltas for each scale
                level with shape (N, num_anchors * 4, H, W)
            centernesses (list[Tensor]): Centerness for each scale
                level with shape (N, num_anchors * 1, H, W)
            gt_bboxes (list[Tensor]): Ground truth bboxes for each image with
                shape (num_gts, 4) in [tl_x, tl_y, br_x, br_y] format.
            gt_labels (list[Tensor]): class indices corresponding to each box
            img_metas (list[dict]): Meta information of each image, e.g.,
                image size, scaling factor, etc.
            gt_bboxes_ignore (list[Tensor] | None): specify which bounding
                boxes can be ignored when computing the loss.

        Returns:
            dict[str, Tensor]: A dictionary of loss components.
        """
        featmap_sizes = [featmap.size()[-2:] for featmap in cls_scores]
        assert len(featmap_sizes) == self.anchor_generator.num_levels

        device = cls_scores[0].device
        anchor_list, valid_flag_list = self.get_anchors(featmap_sizes, img_metas, device=device)
        label_channels = self.cls_out_channels if self.use_sigmoid_cls else 1

        cls_reg_targets = self.get_targets(
            anchor_list,
            valid_flag_list,
            gt_bboxes,
            img_metas,
            gt_bboxes_ignore_list=gt_bboxes_ignore,
            gt_labels_list=gt_labels,
            label_channels=label_channels,
        )
        if cls_reg_targets is None:
            return None

        (
            anchor_list,
            labels_list,
            label_weights_list,
            bbox_targets_list,
            bbox_weights_list,
            valid_label_mask,
            num_total_pos,
            num_total_neg,
        ) = cls_reg_targets

        num_total_samples = reduce_mean(torch.tensor(num_total_pos, dtype=torch.float, device=device)).item()
        num_total_samples = max(num_total_samples, 1.0)

        losses_cls, losses_bbox, loss_centerness, bbox_avg_factor = multi_apply(
            self.loss_single,
            anchor_list,
            cls_scores,
            bbox_preds,
            centernesses,
            labels_list,
            label_weights_list,
            bbox_targets_list,
            valid_label_mask,
            num_total_samples=num_total_samples,
        )

        bbox_avg_factor = sum(bbox_avg_factor)
        bbox_avg_factor = reduce_mean(bbox_avg_factor).item()
        if bbox_avg_factor < EPS:
            bbox_avg_factor = 1
        losses_bbox = list(map(lambda x: x / bbox_avg_factor, losses_bbox))
        return dict(loss_cls=losses_cls, loss_bbox=losses_bbox, loss_centerness=loss_centerness)

    def loss_single(
        self,
        anchors,
        cls_score,
        bbox_pred,
        centerness,
        labels,
        label_weights,
        bbox_targets,
        valid_label_mask,
        num_total_samples,
    ):
        """Compute loss of a single scale level.

        Args:
            cls_score (Tensor): Box scores for each scale level
                Has shape (N, num_anchors * num_classes, H, W).
            bbox_pred (Tensor): Box energies / deltas for each scale
                level with shape (N, num_anchors * 4, H, W).
            anchors (Tensor): Box reference for each scale level with shape
                (N, num_total_anchors, 4).
            labels (Tensor): Labels of each anchors with shape
                (N, num_total_anchors).
            label_weights (Tensor): Label weights of each anchor with shape
                (N, num_total_anchors)
            bbox_targets (Tensor): BBox regression targets of each anchor wight
                shape (N, num_total_anchors, 4).
            num_total_samples (int): Number os positive samples that is
                reduced over all GPUs.

        Returns:
            dict[str, Tensor]: A dictionary of loss components.
        """

        anchors = anchors.reshape(-1, 4)
        cls_score = cls_score.permute(0, 2, 3, 1).reshape(-1, self.cls_out_channels).contiguous()
        bbox_pred = bbox_pred.permute(0, 2, 3, 1).reshape(-1, 4)
        centerness = centerness.permute(0, 2, 3, 1).reshape(-1)
        bbox_targets = bbox_targets.reshape(-1, 4)
        labels = labels.reshape(-1)
        label_weights = label_weights.reshape(-1)
        valid_label_mask = valid_label_mask.reshape(-1, self.cls_out_channels)

        # FG cat_id: [0, num_classes -1], BG cat_id: num_classes
        bg_class_ind = self.num_classes
        pos_inds = ((labels >= 0) & (labels < bg_class_ind)).nonzero().squeeze(1)

        if self.use_qfl:
            quality = label_weights.new_zeros(labels.shape)

        if len(pos_inds) > 0:
            pos_bbox_targets = bbox_targets[pos_inds]
            pos_bbox_pred = bbox_pred[pos_inds]
            pos_anchors = anchors[pos_inds]
            pos_centerness = centerness[pos_inds]

            centerness_targets = self.centerness_target(pos_anchors, pos_bbox_targets)
            pos_decode_bbox_pred = self.bbox_coder.decode(pos_anchors, pos_bbox_pred)
            pos_decode_bbox_targets = self.bbox_coder.decode(pos_anchors, pos_bbox_targets)

            if self.use_qfl:
                quality[pos_inds] = bbox_overlaps(
                    pos_decode_bbox_pred.detach(), pos_decode_bbox_targets, is_aligned=True
                ).clamp(min=1e-6)

            # regression loss
            loss_bbox = self.loss_bbox(
                pos_decode_bbox_pred, pos_decode_bbox_targets, weight=centerness_targets, avg_factor=1.0
            )

            # centerness loss
            loss_centerness = self.loss_centerness(pos_centerness, centerness_targets, avg_factor=num_total_samples)

        else:
            loss_bbox = bbox_pred.sum() * 0
            loss_centerness = centerness.sum() * 0
            centerness_targets = bbox_targets.new_tensor(0.0)

        # Re-weigting BG loss
        if self.bg_loss_weight >= 0.0:
            neg_indices = labels == self.num_classes
            label_weights[neg_indices] = self.bg_loss_weight

        if self.use_qfl:
            labels = (labels, quality)  # For quality focal loss arg spec

        # classification loss
        if isinstance(self.loss_cls, CrossSigmoidFocalLoss):
            loss_cls = self.loss_cls(
                cls_score, labels, label_weights, avg_factor=num_total_samples, valid_label_mask=valid_label_mask
            )
        else:
            loss_cls = self.loss_cls(cls_score, labels, label_weights, avg_factor=num_total_samples)

        return loss_cls, loss_bbox, loss_centerness, centerness_targets.sum()

    def get_targets(
        self,
        anchor_list,
        valid_flag_list,
        gt_bboxes_list,
        img_metas,
        gt_bboxes_ignore_list=None,
        gt_labels_list=None,
        label_channels=1,
        unmap_outputs=True,
    ):
        """Get targets for Detection head.

        This method is almost the same as `AnchorHead.get_targets()`. Besides
        returning the targets as the parent method does, it also returns the
        anchors as the first element of the returned tuple.
        However, if the detector's head loss uses CrossSigmoidFocalLoss,
        the labels_weights_list consists of (binarized label schema * weights) of batch images
        """
        return self.get_atss_targets(
            anchor_list,
            valid_flag_list,
            gt_bboxes_list,
            img_metas,
            gt_bboxes_ignore_list,
            gt_labels_list,
            label_channels,
            unmap_outputs,
        )



"""This is a legacy file that serves to maintain compatibility with the OTX detection framework.

It links to the model_lifecycle enum under 'enums'

# TODO: Remove once https://jira.devtools.intel.com/browse/CVS-67869 is done
"""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .enums.model_lifecycle import ModelLifecycle

__all__ = ["ModelLifecycle"]
